#define _GNU_SOURCE
#include <stdio.h>
#include <stdint.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dlfcn.h>
#include <fcntl.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <time.h>
#include <sys/types.h>
#include <sys/shm.h>

/* standard syscall handlers */
static int (*_open64)(const char *path, int oflag, ...) = NULL;
static ssize_t (*_write)(int fildes, const void *buf, size_t nbyte) = NULL;
static ssize_t (*_read)(int fd, void *buf, size_t count) = NULL;
static int (*_close)(int fd) = NULL;
static off_t (*_lseek64)(int fildes, off_t offset, int whence) = NULL;
static int (*_fstat64)(int filedes, struct stat64 *buf) = NULL;

/* trace log */
static FILE *trace_file = NULL;

#define QCG_GAUSSIAN_HELP "QCG_GAUSSIAN_HELP"
#define QCG_GAUSSIAN_RWF_SUFIX "QCG_GAUSSIAN_RWF_SUFIX"
#define QCG_GAUSSIAN_TRACEFILE "QCG_GAUSSIAN_TRACEFILE"
#define QCG_GAUSSIAN_SESSION_KEY "QCG_GAUSSIAN_SESSION_KEY" /* must be int */
#define QCG_GAUSSIAN_PARTITIONS "QCG_GAUSSIAN_PARTITIONS" /* Syntax: path[:maxMB][,path[:maxMB]]* e.g. /dev/shm:1000,/tmp:1000,/home/plgrid/plgmamonski" */

typedef struct partition_s {
	int fd;
	long long size;
	long long max_size;
	char path[256];
} *partition_t;

typedef struct file_entry_s {
	int fd;
	long long pos;
	long long size;
	int part_count;
	struct partition_s part[5];
} *file_entry_t;

static file_entry_t file = NULL;
static time_t start_time = 0;
static int shmid = -1;
static int session_id = -1;

static void log_msg(const char *pattern, ...);

static int parse_partitions_list(file_entry_t entry)
{
	char *part_list = NULL;
	char *ctxt = NULL;
	char *token = NULL;

	if (getenv(QCG_GAUSSIAN_PARTITIONS) != NULL) {
		part_list = strdup(getenv(QCG_GAUSSIAN_PARTITIONS));
	} else {
		part_list = strdup(".");
	}

	if (!part_list)
		return 0;

	entry->part_count = 0;

	for (token = strtok_r(part_list, ",", &ctxt); token; token = strtok_r(NULL, ",", &ctxt)) {
		char *p_colon = NULL;

		log_msg("parsing partition entry: %s", token);

		if ((p_colon = strchr(token, ':')) != NULL) {
			*p_colon = '\0';
			p_colon++;
		}
	
		snprintf(entry->part[entry->part_count].path, 255, "%s/%d.rwf", token, session_id);	
		entry->part[entry->part_count].fd = -1;
		entry->part[entry->part_count].size = 0;

		if (p_colon) {
			entry->part[entry->part_count].max_size = strtoll(p_colon, (char **)NULL, 10) * (1024LL * 1024LL);
		} else {
			entry->part[entry->part_count].max_size = 1099511627776LL; /* assume one Tera */
		}
		
		entry->part_count++;
	}

	free(part_list);
	
	return entry->part_count;
}

static void log_break_line(void)
{
	if (!trace_file)
		return;

	fprintf(trace_file,"\n");

} 

static void log_call(const char *pattern, ...)
{
	va_list ap;

	if (!trace_file)
		return;

	fprintf(trace_file, "[%d][%6d] ", (int)time(NULL), (int)(time(NULL) - start_time));

	va_start(ap, pattern);
	vfprintf(trace_file, pattern, ap); 
	va_end(ap);

	if (file) 
		fprintf(trace_file, " [pos=%lld,size=%lld] ", file->pos, file->size);

}

static int log_error(int fd, const char *function_name, size_t count)
{
	int save_errno = errno;
	FILE *error_log = fopen("qcg-gaussian.errors", "a");
	
	if (error_log) {
		fprintf(error_log, "time=%d, function = %s, fd=%d, count=%lld, errno=%d, strerror=%s\n", (int)time(NULL), function_name, fd, (long long)count, save_errno, strerror(save_errno));
		fclose(error_log);
	}
	
	return save_errno;
}

static long long log_return(long long return_value)
{

	if (!trace_file)
		return return_value;

        fprintf(trace_file, "    = %lld\n", return_value);
        fflush(trace_file);

	return return_value;
}

static void log_msg(const char *pattern, ...)
{
	va_list ap;

	if (!trace_file)
		return;

	fprintf(trace_file, "[%d][%6d] ", (int)time(NULL), (int)(time(NULL) - start_time));

	va_start(ap, pattern);
	vfprintf(trace_file, pattern, ap); 
	va_end(ap);

	fprintf(trace_file, "\n");
	fflush(trace_file);
}

#define SAVE_SYSCALL(name) do { \
	if (! _ ## name) \
		assert( _ ## name  = dlsym(RTLD_NEXT, #name)); \
} while (0)

static void set_close_on_exec(FILE *f) {
	int fd, oldflags;

	if (!f)
		return;
	if ((fd = fileno(trace_file)) == -1)
		return;

        if ((oldflags = fcntl (fd, F_GETFD, 0)) == -1)
		return;

        oldflags |= FD_CLOEXEC;

        if (fcntl(fd, F_SETFD, oldflags) == -1)
		return;
}

static void help_and_exit() {
	fprintf(stderr, "Set " QCG_GAUSSIAN_HELP " environment variable to get more help\n");
	exit(3);
}

__attribute__((constructor)) void qcg_memcache_init(void) {
	const char *trace_file_path = getenv(QCG_GAUSSIAN_TRACEFILE);

	if (getenv(QCG_GAUSSIAN_HELP)) {
		fprintf(stderr, "QCG File I/O Managament library for Gaussian.\n");
		fprintf(stderr, "Environment variables:\n");
		fprintf(stderr,  QCG_GAUSSIAN_RWF_SUFIX "   - intercept calls to file with the given extension (Defult: '.rwf')\n");
		fprintf(stderr,  QCG_GAUSSIAN_TRACEFILE "   - turns on verbose logging into the given file (Defult: do not produce traces)\n");
		fprintf(stderr,  QCG_GAUSSIAN_PARTITIONS "  - scratch partition list, format: path[:maxMB][,path[:maxMB]]* \n");
		fprintf(stderr, "                           e.g. /dev/shm:1000,/scratch (Defult: './' i.e. only current directory)\n");
		fprintf(stderr,  QCG_GAUSSIAN_SESSION_KEY " - session key to be used as shmid (Mandatory parameter)\n");
		exit(3);
	}

	if (trace_file_path && strlen(trace_file_path)) {
		trace_file = fopen(trace_file_path, "a");
		start_time = time(NULL);

		log_msg("Log opened on: %d (pid=%d,ppid=%d)", (int)start_time, (int)getpid(),(int)getppid());

		set_close_on_exec(trace_file);	
	}

}

__attribute__((destructor)) void qcg_memcache_cleanup(void) {

	if (trace_file) {
		log_msg("Log closed (pid=%d).", getpid());
		fclose(trace_file);
	}
}

int open64(const char *path, int oflag, ...)
{
	int fd;
	static const char *rwf_pattern;
	va_list ap;
	
	SAVE_SYSCALL(open64);
	
	if (!rwf_pattern) {
		rwf_pattern = getenv(QCG_GAUSSIAN_RWF_SUFIX);

		if (!rwf_pattern)
			rwf_pattern = ".rwf";
	}
		
	if (strlen(path) >= strlen(rwf_pattern) && strcmp(path + strlen(path) - strlen(rwf_pattern), rwf_pattern) == 0) {
		log_msg("found match: %s", path); 

		if (oflag & O_CREAT) {
			mode_t mode = 0;
			va_start(ap, oflag);
			mode = va_arg(ap, mode_t);
			log_call("open(%s,%d,%d)", path, oflag, (int)mode);
			fd = _open64(path, oflag, mode);
			va_end(ap);
		} else {
			log_call("open(%s,%d)", path, oflag);
			fd = _open64(path, oflag);
			va_end(ap);
		}
		
		log_return(fd);
		
		if (fd != -1 && shmid == -1) {
			int newmem = 0, i = 0;
			
			if (!getenv(QCG_GAUSSIAN_SESSION_KEY)) {
                                fprintf(stderr, "QCG_GAUSSIAN_SESSION_KEY environment variable not defined\n");
				help_and_exit();
			}

			session_id = strtol(getenv(QCG_GAUSSIAN_SESSION_KEY), (char**)NULL, 10);
			log_msg("Session id = %d", session_id);

			if ((shmid = shmget((key_t)session_id, sizeof(*file), 0 )) == -1) {
				shmid = shmget((key_t)session_id, sizeof(*file), 0600 | IPC_CREAT );
				newmem = 1;
				log_msg("Creating  shared file struct of size: %lld", sizeof(*file));
			} else {
				log_msg("Attaching shared file struct of size: %lld.", sizeof(*file));
			}
			
			if (shmid == -1) {
				perror("shmget failed");
				help_and_exit();
			}

			file = shmat(shmid, (void *)0, 0);
    			
			if (file == (void *)-1) {
				perror("shmat failed");
                                help_and_exit();

    			}
			
			
			if (newmem) {
				file->size = 0;
				file->pos = 0;
				file->fd = fd;

				if (parse_partitions_list(file) == 0) {
					fprintf(stderr, "Failed to read paritions list: %s\n", getenv(QCG_GAUSSIAN_PARTITIONS));
					help_and_exit();
				}	
				
			} else {
				file->pos = 0;
				log_msg("File size = %lld", file->size);
				/* do not set file size */
			}
			/* open all sub-files */
			for (i=0; i < file->part_count; i++) {
				if ((file->part[i].fd = _open64(file->part[i].path, O_RDWR | O_CREAT, 0600)) == -1) {
					fprintf(stderr, "Failed to open sub-file: %s\n",  file->part[i].path);
					help_and_exit();

				}
				
				log_msg("Opened sub-file: %s -> %d", file->part[i].path, file->part[i].fd);
			}
			
		}

		return fd;
	} else { 
		if (oflag & O_CREAT) {
			mode_t mode = 0;
			va_start(ap, oflag);
			mode = va_arg(ap, mode_t);
			fd = _open64(path, oflag, mode);
			va_end(ap);
		} else {
			fd = _open64(path, oflag);
			va_end(ap);
		}
		
		return fd;
	}

}

/* this function is a workaround for Lustre fs */
static ssize_t robust_write(int fd, char *buf, size_t count)
{
	ssize_t bytes_written = 0;
	int tries_left = 3;

	//log_msg("robust_write(%d,...,%lld)", fd, count);

	while (bytes_written != count) {
		ssize_t ret = -1;
		ret = _write(fd, (void*)(buf + bytes_written), count - bytes_written);
		//log_msg("_write ret = %d", ret);

		if (ret == -1) {
			errno = log_error(fd, "write", count);
	
			if (tries_left--) {
				sleep(5 - tries_left);
			} else {
				return -1;
			}
			
		} else if (ret == 0) {
			return bytes_written;
		} else {
			bytes_written += ret;
		}

		if (bytes_written != count) {
			log_msg("WARNING: only %lld of %lld bytes written (fd=%d)", bytes_written, count, fd);
		}
	}

	return bytes_written;
}

static ssize_t write_zeros(int fd,  size_t count) {
	char zerobuf[4096];
	ssize_t bwritten = 0;

	log_msg("# write_zeros(%d,%lld)", fd, count); 
	memset(zerobuf, 0, sizeof(zerobuf));

	while (bwritten != count) {
		ssize_t ret = robust_write(fd, zerobuf, count > sizeof(zerobuf) ? sizeof(zerobuf) : count);
		
		if (ret == -1)
			return bwritten;

		bwritten += ret;	
	}

	log_msg("# write_zeros -> %lld", bwritten); 
	return bwritten;
}

static ssize_t internal_write(long long pos, char *buf, size_t count)
{
	int i = 0;
	long long offset = 0;
	long long bytes_written1 = 0;
	long long bytes_written2 = 0;

	assert(count > 0);	

	log_break_line();
	log_msg("# internal_write(%lld,%p,%lld)", pos, buf, count);
	while (i < file->part_count &&  
			((offset + file->part[i].size < pos && file->part[i].size) ||
			 (file->part[i].max_size == 0) || /* boundary condition - partition was fully occupied */
			 (offset + file->part[i].max_size == pos) /* would not write anything in the file - max_size exceded */
				)) {
		log_msg("# part[%d].size = %lld", i, file->part[i].size);
		offset += file->part[i].size;
		i++;
	}


	if (i == file->part_count) /*we are beyond end of file */ {
		log_msg("# ERROR: Not enough partitions (total partitions=%d)\n", file->part_count);
		fprintf(stderr, "QCG-ERROR: Not enough partitions (total partitions=%d, file size=%lld, file pos=%lld)\n", file->part_count, file->size, file->pos);
		return -1;
	}
	log_msg("# Offset: global=%lld, local=%lld", offset, pos - offset);
	/* translate global offset to local offset */
	offset = pos - offset;

	/* now try to write, there are 2 cases */
	/* 1. max file size is smaller than bytes written */
	if (offset + count > file->part[i].max_size) {
		long long chunk1_count = file->part[i].max_size - offset;
		long long chunk2_count = count - chunk1_count;

		log_msg("# 1. Splitting write into two chunks of size %lld and %lld", chunk1_count, chunk2_count);
		/* split write into two calls */
		if ((bytes_written1 = internal_write(pos, buf, chunk1_count)) != chunk1_count)
			return bytes_written1; /*we can only continue if we have already written what we want */

		log_msg("# 1. bytes_written1 = %lld", bytes_written1);

		if ((bytes_written2 = internal_write(pos + chunk1_count, (buf != NULL ? buf + chunk1_count : NULL), chunk2_count)) == -1)
			return bytes_written1; /* Fail !!! Return what we have written already */
		
		log_msg("# 1. bytes_written2 = %lld", bytes_written1);

		return bytes_written1 + bytes_written2;
	} else {
		/* seek into file */
		log_msg("# 2. Local seek: fd=%d, offset=%lld", file->part[i].fd, offset);
		if (lseek(file->part[i].fd, offset, SEEK_SET) != offset)
			return -1;

		/* in theory we should write all at once */
		if (buf != NULL) {
			bytes_written1 = robust_write(file->part[i].fd, buf, count);
		} else {
			bytes_written1 = write_zeros(file->part[i].fd, count);
		}

		if (bytes_written1 > 0 && file->part[i].size < (offset + bytes_written1) ) {
			/* we have written something, update file size if needed */
			file->part[i].size = offset + bytes_written1;
		}

		log_msg("# 2. bytes_written1 = %lld", bytes_written1);

		if (bytes_written1 == count)
			return bytes_written1; /*we are done */

		/* but... */
		
		if (bytes_written1 == -1 && errno != ENOSPC) /*we may get some IO error */
			return -1;

		if ((bytes_written1 == -1 && errno == ENOSPC) || (bytes_written1 == 0)) {
			file->part[i].max_size = file->part[i].size; /* update "high water mark" */

			return internal_write(pos, buf, count); /* the next call should try with next partition */
		} else if (bytes_written1 < count) { /* we have written something */

			assert(offset + bytes_written1 >= file->part[i].size); /* we assume that we are extending file */
			file->part[i].size = offset + bytes_written1;

			if ((bytes_written2 = internal_write(pos + bytes_written1, (buf != NULL ? buf + bytes_written1: NULL), count - bytes_written1)) == -1)
				return bytes_written1;

			return bytes_written1 + bytes_written2;
		} else {
			assert(0);
		}
	}
}

ssize_t write(int fd, const void *buf, size_t count)
{
	SAVE_SYSCALL(write);

	
	if (file && file->fd == fd) {
		
		log_call("write(%d, %p, %d)",fd, buf, count);
		
		if ((count = internal_write(file->pos, (char*)buf, count)) == -1)	
			return -1;

		file->pos += count;
		if (file->pos > file->size)
			file->size = file->pos;

		return log_return(count);
	}
	
	return _write(fd, buf, count);
}

/* this function is a workaround for Lustre fs */
static ssize_t robust_read(int fd, char *buf, size_t count)
{
	ssize_t bytes_read = 0;
	int tries_left = 3;

	//log_msg("robust_read(%d,..,%lld)", fd, count);

	while (bytes_read != count) {
		ssize_t ret = -1;
		ret = _read(fd, (void*)(buf + bytes_read), count - bytes_read);

		//log_msg("ret = %d", ret);

		if (ret == -1) {
			errno = log_error(fd, "read", count);
	
			if (tries_left--) {
				sleep(5 - tries_left);
			} else {
				return -1;
			}
			
		} else if (ret == 0) {
			return bytes_read;
		} else {
			bytes_read += ret;
		}

		if (bytes_read != count) {
			log_msg("WARNING: only %lld of %lld bytes read (fd=%d)", bytes_read, count, fd);
		}
	}

	//log_msg("bytes_read = %d", bytes_read);
	return bytes_read;
}

static ssize_t internal_read(long long pos, char *buf, size_t count)
{
	int i = 0;
	long long offset = 0;

	log_break_line();
        log_msg("# internal_read(%lld,%p,%lld)", pos, buf, count);
	
	while (i < file->part_count && (offset + file->part[i].size <= pos)) {
		log_msg("# part[%d].size = %lld", i, file->part[i].size);
		offset += file->part[i].size;
		i++;
	}

	log_msg("# Offset: global=%lld, local=%lld", offset, pos-offset);

	if (offset == file->size)
		return 0; /* EOF */
	else if (offset > file->size || i == file->part_count)
		return -1;

	/* translate global offset to local offset */
	offset = pos - offset;

	/* seek into file */
	log_msg("# Local seek: fd=%d, offset=%lld", file->part[i].fd, offset);
	if (lseek(file->part[i].fd, offset, SEEK_SET) != offset)
		return -1;

	/* */	
	if (offset + count <= file->part[i].size) {
		return robust_read(file->part[i].fd, buf, count);
	} else {
		long long local_count = -1;
		long long bytes_read1 = -1;
		long long bytes_read2 = -1;
 		
		local_count = file->part[i].size - offset;

		if ((bytes_read1 = robust_read(file->part[i].fd, buf, local_count)) != local_count) /*we expect to read local_count */
			return bytes_read1; 
		

		if ((bytes_read2 = internal_read(pos + local_count, buf + local_count, count - local_count)) == -1)
			return -1;

		return bytes_read1 + bytes_read2;
	}
}


ssize_t read(int fd, void *buf, size_t count)
{
	SAVE_SYSCALL(read);

	if (file && file->fd == fd) {
        	log_call("read(%d, ..., %d)", fd, count);
		
		if ((count = internal_read(file->pos, (char*)buf, count)) == -1)
			return log_return(-1);
	
		file->pos += count;

		return log_return(count);
	}
	
	return _read(fd, buf, count);
}

int close(int fd)
{
	SAVE_SYSCALL(close);
	
	if (file && file->fd == fd) {
		int i = 0;
		log_call("close(%d)", fd);
		file->pos = 0;
		for (i = 0; i < file->part_count; i++) {
			_close(file->part[i].fd);
			file->part[i].fd = -1;
		}

		return log_return(_close(fd));
	}
	
	return _close(fd);
}



off_t lseek64(int fd, off_t offset, int whence)
{
	SAVE_SYSCALL(lseek64);
	SAVE_SYSCALL(write); /* seek initalize empty space with zeros, thus we need save the write call also here */


	if (file && file->fd == fd) {

		log_call("lseek(%d, %lld, %s)", fd, (long long)offset, (whence == SEEK_SET ? "SEEK_SET" : (whence == SEEK_CUR ? "SEEK_CUR" : "SEEK_END")));

		if (whence == SEEK_SET) {
			file->pos = offset;
		} else if (whence == SEEK_CUR) {
			file->pos += offset;
		} else if (whence == SEEK_END) {
			file->pos = file->size + offset;
		} else {
			assert(0);
		}
		
		if (file->pos > file->size) { /*write zeros, this make sure that the "gap space" would be actually available for further writes*/
			long long count = file->pos - file->size;

			if (internal_write(file->size, NULL, count) != count)
				return log_return(-1);
			
			file->size = file->pos; /*invariant:  file->size + count = file->size - file->size + file->pos = file->pos */
		}

		return log_return(file->pos);
	}
	
        return _lseek64(fd, offset, whence);
}

int fstat64(int fd, struct stat64 *buf)
{
	SAVE_SYSCALL(fstat64);
	
	
	if (file && file->fd == fd) {
		int ret = _fstat64(fd, buf);

		log_call("fstat64(%d", fd);

		//cheat about size
		buf->st_size = file->size;
                return log_return(ret); 
        }


	return _fstat64(fd, buf);
}



