#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>

#include <signal.h>
#include "pty_fork.h"
#include "ssl_util.h"
#include "socket_util.h"
#include "debug.h"


#define DEFAULT_PORT "20100"
#define DEFAULT_APP "/bin/bash"

#define DEFAULT_SHELL_ROWS  25
#define DEFAULT_SHELL_COLS  80

#define BUF_SIZE 16 * 1024


/*****************************************************************************/


#define _min(a, b)		((a) < (b) ? (a) : (b))


/*****************************************************************************/


struct args_data {
	char host[128];
	char port[24];
	char app[128];
	char *const *app_args;

	int rows, cols;
};


void help(void);

void child(struct args_data *args);

int write_fd(int fd, int count, char *buf);

int stream_pty(int outputFd, int ptyFd, struct ssl_data *ssl);

int setup_args(struct args_data *args, int argc, char **argv);

void setup_debug(void);

int setup_signals(void);

static void sig_handler(int sig, siginfo_t *siginfo, void *context);
	
/*****************************************************************************/


int main(int argc, char **argv) {
	struct winsize slaveWS;
	int childPid;
	int outputFd, ptyFd;
	struct ssl_data ssl;
	struct args_data args;

	setup_debug();

	if (setup_args(&args, argc, argv))
		exit(1);

    signal(SIGPIPE, SIG_IGN);
    
	slaveWS.ws_row = args.rows;
	slaveWS.ws_col = args.cols;

	if (setup_signals())
		exit(2);
	
	childPid = ptyFork(&ptyFd, NULL, 0, NULL, &slaveWS);
	if (childPid == -1)
		ERR_EXIT("ptyFork");

	if (childPid == 0)   /* Child: execute a shell on pty slave */
		child(&args);

	if ((outputFd = connectTo(args.host, args.port)) < 0)
		exit(2);

	if (find_proxy_cert(ssl.cert_file, sizeof(ssl.cert_file)))
		exit(3);

	if (find_ca_path(ssl.ca_certs_path, sizeof(ssl.ca_certs_path)))
		exit(4);

	if (ssl_setup(outputFd, &ssl)) {
		closeSocket(outputFd);
		exit(3);
	}

	stream_pty(outputFd, ptyFd, &ssl);

	return 0;
}


/*****************************************************************************/


int setup_args(struct args_data *args, int argc, char **argv) {
	const char *session_contact;
	int i, arg_next;

	session_contact = getenv("QCG_INT_SESSION_CONTACT");
	LOG(L_DEBUG, "session contact variable [%s]", session_contact);

	arg_next = 1;

	memset(args->host, 0, sizeof(args->host));
	memset(args->port, 0, sizeof(args->port));
	memset(args->app, 0, sizeof(args->app));

	if (!session_contact) {
		if (argc < 3) {
			LOG(L_ERROR, "error: wrong syntax\n\n");
			help();
			return 1;
		}

		strncpy(args->host, argv[1], sizeof(args->host));
		strncpy(args->port, argv[2], sizeof(args->port));
		LOG(L_DEBUG, "getting host & port from args (%s:%s)", args->host, args->port);
		arg_next = 3;
	} else {
		if (index(session_contact, ':')) {
			strncpy(args->host, session_contact, _min(
						index(session_contact, ':') - session_contact,
						sizeof(args->host)));
			strncpy(args->port, index(session_contact, ':') + 1,
					sizeof(args->port));
			LOG(L_DEBUG, "parsing session contact variable with host & port");
		} else {
			LOG(L_DEBUG, "session contact contains only host");
			strncpy(args->host, session_contact, sizeof(args->host));
			strncpy(args->port, DEFAULT_PORT, sizeof(args->port));
		}
	}
	
	LOG(L_DEBUG, "final host & port (%s:%s)", args->host, args->port);

	if (arg_next < argc) {
		strncpy(args->app, argv[arg_next++], sizeof(args->app));
	} else {
		strncpy(args->app, DEFAULT_APP, sizeof(args->app));
	}

	if (arg_next < argc)
		args->app_args = &argv[arg_next];
	else
		args->app_args = NULL;

	args->rows = DEFAULT_SHELL_ROWS;
	args->cols = DEFAULT_SHELL_COLS;

	if (getenv("QCG_SHELL_ROWS")) {
		LOG(L_DEBUG, "geting # of rows from env (%s)", getenv("QCG_SHELL_ROWS"));
		args->rows = atoi(getenv("QCG_SHELL_ROWS"));
	}

	if (getenv("QCG_SHELL_COLUMNS")) {
		LOG(L_DEBUG, "geting # of cols from env (%s)", getenv("QCG_SHELL_COLUMNS"));
		args->cols = atoi(getenv("QCG_SHELL_COLUMNS"));
	}

	LOG(L_DEBUG, "final size of console (%d x %d)", args->rows, args->cols);

	LOG(L_INFO, "app [%s]", args->app);
	if (args->app_args) {
		LOG(L_INFO, " with arguments:");

		for (i = 0; args->app_args[i]; ++i)
			LOG(L_INFO, " '%s'", args->app_args[i]);
	}

	return 0;
}


int stream_pty(int outputFd, int ptyFd, struct ssl_data *ssl) {
	char buf[BUF_SIZE];
	char err_buf[512];
	fd_set inFds;
	ssize_t numRead;
	int maxFd;
	int res;

	maxFd = (ptyFd > outputFd ? ptyFd : outputFd) + 1;
	for (;;) {
		FD_ZERO(&inFds);
		FD_SET(ptyFd, &inFds);
		FD_SET(outputFd, &inFds);

		if (select(maxFd, &inFds, NULL, NULL, NULL) == -1) {
			LOG(L_ERROR, "failed to select file descriptors");
			return -1;
		}

		if (FD_ISSET(outputFd, &inFds)) {   /* remote --> pty */
//			LOG(L_DEBUG, "data available on ssl socket", numRead);

			errno = 0;
			numRead = ssl_read(ssl, BUF_SIZE, buf);
			int err_no = errno;
			if (numRead <= 0) {
				/*
				BIO *bio;
				BUF_MEM *mem;
				if ((bio = BIO_new(BIO_s_mem()))) {
					        ERR_print_errors(bio);
					        BIO_get_mem_ptr(bio, &mem);
							LOG(L_ERROR, "bio errors: %s", mem->data);
					        BIO_free_all(bio);
			    } 
				*/

				int err_err;
				char err_buf[512];
				while ((err_err = ERR_get_error()) != 0) {
					ERR_error_string(err_err, err_buf);
					LOG(L_ERROR, "err_get_error: %s", err_buf);
			    }

				int ssl_err = SSL_get_error(ssl->ssl, numRead); 
				LOG(L_ERROR, "ssl read result(%d), error num(%d), errno(%d)", numRead, ssl_err, err_no);
				
				if (ssl_err == SSL_ERROR_WANT_READ || ssl_err == SSL_ERROR_WANT_WRITE) {
					LOG(L_ERROR, "ssl_read WANT_READ/WANT_WRITE (%d)", ssl_err);
				}

				ERR_error_string_n(SSL_get_error(ssl->ssl, numRead), err_buf, sizeof(err_buf));
				LOG(L_ERROR, "ssl read error: %s", err_buf);
				return 0;
			} else {
//				LOG(L_DEBUG, "read %d bytes from ssl socket", numRead);
			}

			if (write_fd(ptyFd, numRead, buf) < 0) {
				LOG(L_ERROR, "failed to write to pty: %s", strerror_r(errno, err_buf, sizeof(err_buf)));
				return -1;
			}
		}

		if (FD_ISSET(ptyFd, &inFds)) {      /* pty --> remote */
//			LOG(L_DEBUG, "data available on pty socket");

			numRead = read(ptyFd, buf, BUF_SIZE);
			if (numRead <= 0) {
				if (numRead < 0)
					LOG(L_ERROR, "failed to read from pty: %s", strerror_r(errno, err_buf, sizeof(err_buf)));
				return 0;
			} else {
//				LOG(L_DEBUG, "read %d bytes from pty socket", numRead);
			}

			if ((res = ssl_write(ssl, numRead, buf)) <= 0) {
				int err_code = SSL_get_error(ssl->ssl, res);
				
				ERR_error_string_n(err_code, err_buf, sizeof(err_buf));
				
				if (err_code == SSL_ERROR_WANT_READ || err_code == SSL_ERROR_WANT_WRITE) {
					LOG(L_ERROR, "ssl_write WANT_READ/WANT_WRITE (%d)", err_code);
				}
				
				LOG(L_ERROR, "failed to write data to remote peer: %s", err_buf);
				return -1;
			} else {
//				LOG(L_DEBUG, "wrote %d bytes to ssl socket", res);
			}
			
			if (res < numRead) {
				LOG(L_ERROR, "not all data has been writen (%d vs %ld)", res, numRead);
			}
		}
	}

	return 0;
}


int write_fd(int fd, int count, char *buf) {
	int writen;

	while (count > 0) {
		if ((writen = write(fd, buf, count)) < 0)
			return -1;

		count -= writen;
		buf += writen;
	}

	return 0;
}


void help(void) {
	printf("\tinteractive [host port] [arguments ...]\n\n");
}


void child(struct args_data *args) {
	execv(args->app, args->app_args);
	ERR_EXIT("execv");      /* If we get here, something went wrong */
}


void setup_debug(void) {
	FILE *file;

	file = fopen("debug.log", "wt");
	if (file)
		log_set_output(file);

	log_set_level(L_DEBUG);
}


int setup_signals(void) {
	struct sigaction act;
	
	memset(&act, 0, sizeof(act));
	act.sa_sigaction = &sig_handler;
	act.sa_flags = SA_SIGINFO;
	if (sigaction(SIGTERM, &act, NULL) < 0) {
		perror("sigaction");
		return 1;
	}
	
	return 0;
}


static void sig_handler(int sig, siginfo_t *siginfo, void *context) {
	fprintf(stderr, "caught signal(%d) pid(%ld), uid(%ld)", sig, (long)siginfo->si_pid, (long)siginfo->si_uid);
	LOG(L_INFO, "caught signal(%d) pid(%ld), uid(%ld)", sig, (long)siginfo->si_pid, (long)siginfo->si_uid);
}

/*****************************************************************************/

