#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <time.h>

#include "ssl_util.h"
#include "validator.h"
#include "debug.h"


/******************************************************************************/

static void debug_cert(X509 *cert);

static void debug_chain(STACK_OF(X509) *chain);

static const char* ssl_last_error(void);

static const char* get_ssl_error(unsigned long err);

static int accept_cb(int preverify_ok, X509_STORE_CTX *ctx);

/******************************************************************************/


int ssl_setup(int fd, struct ssl_data *ssl) {
	STACK_OF(X509) *chain;
	X509 *cert;
	char client_dn[256];
	char server_dn[256];
	int nonproxy_idx;

	ssl->ctx = NULL;
	chain    = NULL;

	/* initialize openssl */
	SSL_load_error_strings();
	ERR_load_BIO_strings();
	SSL_library_init();
	OpenSSL_add_all_algorithms();

//	ssl->ctx = SSL_CTX_new(SSLv23_client_method());
//	ssl->ctx = SSL_CTX_new(SSLv3_client_method());
//	ssl->ctx = SSL_CTX_new(TLSv1_client_method());
	ssl->ctx = SSL_CTX_new(SSLv3_client_method());
	
	int vflags = 0;
	vflags |= X509_V_FLAG_ALLOW_PROXY_CERTS;
	vflags |= X509_PURPOSE_ANY;
	vflags |= X509_V_FLAG_IGNORE_CRITICAL;
//	vflags |= X509_V_FLAG_X509_STRICT;

	X509_STORE_set_flags(SSL_CTX_get_cert_store(ssl->ctx), vflags);
	/* end setup ssl context parameters */
	/****/

//	SSL_CTX_set_options(ssl->ctx, SSL_OP_ALL);
//	SSL_CTX_set_options(ssl->ctx, SSL_OP_TLS_ROLLBACK_BUG | SSL_OP_SINGLE_DH_USE);
//	SSL_CTX_set_options(ssl->ctx, SSL_OP_NO_SSLv2 | SSL_OP_NO_SSLv3 | SSL_OP_NO_SESSION_RESUMPTION_ON_RENEGOTIATION | SSL_OP_NO_TICKET | SSL_OP_ALLOW_UNSAFE_LEGACY_RENEGOTIATION); 
//	SSL_CTX_set_options(ssl->ctx, SSL_OP_NO_TLSv1 | SSL_OP_NO_SSLv2); 

	/****/
	/* begin setup ca, cert & key */
	if (! SSL_CTX_load_verify_locations(ssl->ctx, NULL, ssl->ca_certs_path)) {
		LOG(L_WARN, "failed to set ca certs file [%s]: %s", ssl->ca_certs_path,
				ssl_last_error());
	}

	if (SSL_CTX_use_certificate_chain_file(ssl->ctx, ssl->cert_file) != 1) {
		LOG(L_WARN, "failed to load certificate '%s': %s", ssl->cert_file,
				ssl_last_error());
	}
	
	if (SSL_CTX_use_PrivateKey_file(ssl->ctx, ssl->cert_file, SSL_FILETYPE_PEM)
			!= 1) {
		LOG(L_ERROR, "failed to load key '%s': %s\n", ssl->cert_file,
				ssl_last_error());
		return 1;
	}

	LOG(L_DEBUG, "certificate '%s' successfully loaded", ssl->cert_file);

	if (SSL_CTX_check_private_key(ssl->ctx) != 1) {
		LOG(L_ERROR, "private key '%s' not correct: %s\n", ssl->cert_file,
				ssl_last_error());
		return 2;
	}

	LOG(L_DEBUG, "private key '%s' successfully loaded", ssl->cert_file);

	/* begin ssl handshake */
	if (!(ssl->bio = BIO_new(BIO_s_socket()))) {
		LOG(L_ERROR, "failed to setup ssl bio");
		return 3;
	}

	if (!(ssl->ssl = SSL_new(ssl->ctx))) {
		LOG(L_ERROR, "failed to create new ssl session");
		return 4;
	}

	BIO_set_fd(ssl->bio, fd, BIO_NOCLOSE);
	SSL_set_bio(ssl->ssl, ssl->bio, ssl->bio);
	SSL_set_mode(ssl->ssl, SSL_MODE_AUTO_RETRY);
	SSL_set_verify(ssl->ssl, SSL_VERIFY_PEER, accept_cb);

	LOG(L_DEBUG, "performing SSL handshake");

	if (SSL_connect(ssl->ssl) <= 0) {
		LOG(L_ERROR, "failed to connect to server: %s", ssl_last_error());
		return 5;
	}

	LOG(L_DEBUG, "SSL connection established");
	/* end ssl handshake */

	/* client dn */
	cert = SSL_get_certificate(ssl->ssl);
	if (cert == NULL) {
		LOG(L_ERROR, "failed to get local certificate");
		return 6;
	}
	X509_NAME_oneline(X509_get_subject_name(cert), client_dn, sizeof(client_dn));
	X509_free(cert);
	LOG(L_DEBUG, "local certificate dn [%s]", client_dn);

	chain = ssl->ctx->extra_certs;
	if (chain != NULL && sk_X509_num(chain) > 0) {
		LOG(L_DEBUG, "local certificate chain");
		debug_chain(chain);

		nonproxy_idx = find_first_non_proxy(chain);
		if (nonproxy_idx >= 0) {
			X509_NAME_oneline(X509_get_subject_name(sk_X509_value(chain,
							nonproxy_idx)), client_dn, sizeof(client_dn));
		}
	}
	/* end setup ca, cert & key */
	/****/


	/** begin verify peer certificate */
	SSL_get_verify_result(ssl->ssl);

	/* peer certificate dn */
	STACK_OF(X509) *peer_chain = SSL_get_peer_cert_chain(ssl->ssl);

	if (peer_chain == NULL || sk_X509_num(peer_chain) < 1) {
		LOG(L_ERROR, "failed to get remote peer cert chain");
		return 8;
	}

	if (validate_chain(ssl, peer_chain))
		return 8;

	LOG(L_DEBUG, "remote peer certificate chain");
	debug_chain(peer_chain);

	nonproxy_idx = find_first_non_proxy(peer_chain);
	if (nonproxy_idx < 0) {
		LOG(L_ERROR, "failed to find non proxy in peer chain");
		return 8;
	}

	X509_NAME_oneline(X509_get_subject_name(sk_X509_value(peer_chain,
					nonproxy_idx)), server_dn, sizeof(server_dn));

	/* compare dn's */
	if (strcmp(server_dn, client_dn)) {
		LOG(L_ERROR, "wrong server certificate: peer dn (%s) doesn't match "
				"client dn (%s)", server_dn, client_dn);
		return 9;
	}

	LOG(L_DEBUG, "server_dn(%s)\nclient_dn(%s)", server_dn, client_dn);

	return 0;
}


int ssl_read(struct ssl_data *ssl, int count, char *buf) {
  return SSL_read(ssl->ssl, buf, count);
}


int ssl_write(struct ssl_data *ssl, int count, char *buf) {
  return SSL_write(ssl->ssl, buf, count);
}


int find_proxy_cert(char *buffer, size_t len) {
	unsigned int uid;
	char proxyName[256];
	struct stat st;

	uid = getuid();

	if (getenv("X509_USER_PROXY")) {
		strncpy(proxyName, getenv("X509_USER_PROXY"), sizeof(proxyName));
	} else {
		snprintf(proxyName, sizeof(proxyName), "/tmp/x509up_u%d", uid);
	}

	LOG(L_DEBUG, "looking for user (%d) proxy at '%s'", uid, proxyName);

	if (stat(proxyName, &st)) {
		LOG(L_ERROR, "proxy file '%s' doesn't exists", proxyName);
		return 1;
	}

	if (st.st_uid != uid) {
		LOG(L_ERROR, "proxy file '%s' contains wrong user id '%d'"
				" (should be '%d')\n", st.st_uid, uid);
		return 2;
	}

	strncpy(buffer, proxyName, len);

	return 0;
}


int find_ca_path(char *buffer, size_t len) {
	struct stat st;

	if (!buffer || len < 1) {
		LOG(L_WARN, "no buffer for ca path");
		return 1;
	}

	if (getenv("X509_CERT_DIR")) {
		strncpy(buffer, getenv("X509_CERT_DIR"), len);
	} else {
		strncpy(buffer, "/etc/grid-security/certificates", len);
	}

	LOG(L_DEBUG, "ca path set to [%s]", buffer);

	if (stat(buffer, &st)) {
		LOG(L_ERROR, "ca path [%s] doesn't exists", buffer);
		return 1;
	}

	if (!S_ISDIR(st.st_mode)) {
		LOG(L_ERROR, "ca path [%s] is not directory", buffer);
		return 1;
	}

	return 0;
}


/******************************************************************************/


static void debug_chain(STACK_OF(X509) *chain) {
	int i;

	if (chain) {
		LOG(L_DEBUG, "certificate chain contains #%d elements",
				sk_X509_num(chain));
		for (i = 0; i < sk_X509_num(chain); ++i) {
			X509 *cert = sk_X509_value(chain, i);

			LOG(L_DEBUG, "%d:", i);
			debug_cert(cert);
		}
	}
}


static void debug_cert(X509 *cert) {
	if (cert) {
		char subject[256], issuer[256];

		X509_NAME_oneline(X509_get_subject_name(cert), subject,
				sizeof(subject));
		X509_NAME_oneline(X509_get_issuer_name(cert), issuer,
				sizeof(issuer));

		LOG(L_DEBUG, " subject [%s]", subject);
		LOG(L_DEBUG, " issuer [%s]", issuer);
	}
}


static const char* ssl_last_error(void) {
	return get_ssl_error(ERR_peek_last_error());
}


static const char* get_ssl_error(unsigned long err) {
	static char buffer[256];

	ERR_error_string_n(err, buffer, sizeof(buffer));
	return buffer;
}


static int accept_cb(int err_no, X509_STORE_CTX *ctx) {
	int err, err_depth;

	err = X509_STORE_CTX_get_error(ctx);
	err_depth = X509_STORE_CTX_get_error_depth(ctx);

	LOG(L_DEBUG, "verify_cert_cb (err_no(%d), err(%d), err_depth(%d))", err_no,
			err, err_depth);

	return 1;
}


/******************************************************************************/
