#include "ssl_util.h"
#include "validator.h"
#include "debug.h"


/******************************************************************************/


static int cert_is_proxy(X509 *cert);


static void list_proxies(STACK_OF(X509) *chain);


static int verify_cert_cb(int preverify_ok, X509_STORE_CTX *ctx);


static int validate_cert_chain(struct ssl_data *ssl, X509 *cert,
		STACK_OF(X509) *cert_chain);


static unsigned long verify_proxy(STACK_OF(X509) *chain, int start_pos,
		int end_pos);


static time_t grid_asn1TimeToTimeT(char *asn1time, size_t len);


/******************************************************************************/


int find_first_non_proxy(STACK_OF(X509) *chain) {
	int i;

	for (i = 0; i < sk_X509_num(chain); ++i) {
		X509 *cert = sk_X509_value(chain, i);

		if (!cert_is_proxy(cert))
			return i;
	}

	return -1;
}



int validate_chain(struct ssl_data *ssl, STACK_OF(X509) *chain) {
	int nonproxy_pos;
	int ret;

	nonproxy_pos = find_first_non_proxy(chain);

	list_proxies(chain);

	/* validate certificates */
	if (nonproxy_pos > 0) {
		X509 *top_cert = sk_X509_value(chain, nonproxy_pos);
		if ((ret = validate_cert_chain(ssl, top_cert, chain)))
			return ret;
	} else {
		LOG(L_ERROR, "lack of non self signed certificate in chain");
		return X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY;
	}

	/* validate proxies */
	if (nonproxy_pos > 0) {
		LOG(L_DEBUG, "validating #%d proxies", nonproxy_pos);
    
		if ((ret = verify_proxy(chain, 0, nonproxy_pos)))
			return ret;

	}

	return 0;
}


/******************************************************************************/


static void list_proxies(STACK_OF(X509) *chain) {
	char subject[128];
	int i;

	for (i = 0; i < sk_X509_num(chain); ++i) {
		X509 *cert = sk_X509_value(chain, i);

		X509_NAME_oneline(X509_get_subject_name(cert), subject,
				sizeof(subject));

		if (cert_is_proxy(cert)) {
			LOG(L_DEBUG, "[%s] IS proxy", subject);
		} else {
			LOG(L_DEBUG, "[%s] IS NOT proxy", subject);
		}
	}
}


static int cert_is_proxy(X509 *cert) {
	int  i;
	char s[80];
	X509_EXTENSION *ex;
	char dn[128], issuer[128];
	int dn_len, issuer_len;
	static const char *PROXYCERTINFO_OID = "1.3.6.1.5.5.7.1.14";
	static const char *OLD_PROXYCERTINFO_OID = "1.3.6.1.4.1.3536.1.222";


	for (i = 0; i < X509_get_ext_count(cert); ++i) {
		ex = X509_get_ext(cert, i);
		
		if (X509_EXTENSION_get_critical(ex)) {
			OBJ_obj2txt(s, sizeof(s), X509_EXTENSION_get_object(ex), 1);

			LOG(L_DEBUG, "Critical extension found: %s", s);

			if (strcmp(s, PROXYCERTINFO_OID) == 0) return 1;
			if (strcmp(s, OLD_PROXYCERTINFO_OID) == 0) return 1;
		}
	}

    X509_NAME_oneline(X509_get_subject_name(cert), dn, sizeof(dn));
    X509_NAME_oneline(X509_get_issuer_name(cert), issuer, sizeof(issuer));

    dn_len     = strlen(dn);
    issuer_len = strlen(issuer);
    
    if (issuer_len < dn_len && !strncmp(dn, issuer, issuer_len)) {
    	return 1;
    }
    
    return 0;
}


static int verify_cert_cb(int err_no, X509_STORE_CTX *ctx) {
	char buf[256];
	X509 *err_cert;
	STACK_OF(X509) *chain;
	int err, err_depth, chain_l;
	time_t now;
	
	err_cert = X509_STORE_CTX_get_current_cert(ctx);
	err = X509_STORE_CTX_get_error(ctx);
	err_depth = X509_STORE_CTX_get_error_depth(ctx);

	chain = X509_STORE_CTX_get_chain(ctx);
	chain_l = sk_X509_num(chain);

	X509_NAME_oneline(X509_get_subject_name(err_cert), buf, 256);

	LOG(L_DEBUG, "verify_cert_cb (err_no(%d), err(%d), err_depth(%d))", err_no,
			err, err_depth);

	if (err == X509_V_ERR_CRL_NOT_YET_VALID)
		return 1;

	if (err == X509_V_ERR_UNABLE_TO_GET_CRL)
		return 1;

	if (err == X509_V_ERR_CERT_NOT_YET_VALID ||
			err == X509_V_ERR_CERT_HAS_EXPIRED) {
		time(&now);
        if (now > grid_asn1TimeToTimeT((char*)ASN1_STRING_data(
						X509_get_notBefore(err_cert)),0) &&
		    now < grid_asn1TimeToTimeT((char*)ASN1_STRING_data(
					X509_get_notAfter(err_cert)),0))
			return 1;
	}

	if (!err_no)
		LOG(L_DEBUG, "verify error: num=%d:%s:depth=%d/%d:%s", err,
				X509_verify_cert_error_string(err), err_depth, chain_l, buf);

	return err_no;
}


static int validate_cert_chain(struct ssl_data *ssl, X509 *cert,
		STACK_OF(X509) *cert_chain) {
	X509_STORE *store = NULL;
	X509_LOOKUP *lookup = NULL;
	X509_STORE_CTX *verify_ctx = NULL;
	int vflags = 0;
	char cert_dn[128], issuer_dn[128];

	LOG(L_DEBUG, "validating cert chain");

	if (!(store = X509_STORE_new())) {
		LOG(L_ERROR, "failed to create store");
		return 1;
	}

	X509_STORE_set_verify_cb_func(store, verify_cert_cb);
	LOG(L_DEBUG, "ca certs path (%s)", ssl->ca_certs_path);
	if (X509_STORE_load_locations(store, NULL, ssl->ca_certs_path) != 1) {
		LOG(L_ERROR, "failed to setup ca path");
		return 2;
	}

	if (X509_STORE_set_default_paths(store) != 1) {
		LOG(L_ERROR, "failed to set default paths");
		return 3;
	}

	if (!(lookup = X509_STORE_add_lookup(store, X509_LOOKUP_hash_dir()))) {
		LOG(L_ERROR, "failed to create hash dir lookup");
		return 4;
	}

	if (!(X509_LOOKUP_add_dir(lookup, ssl->ca_certs_path, X509_FILETYPE_PEM))) {
		LOG(L_ERROR, "failed to setup lookup");
		return 5;
	}

	vflags |= X509_V_FLAG_ALLOW_PROXY_CERTS;
	vflags |= X509_PURPOSE_ANY;
	vflags |= X509_V_FLAG_IGNORE_CRITICAL;
//	vflags |= X509_V_FLAG_CRL_CHECK;
//	vflags |= X509_V_FLAG_CRL_CHECK_ALL;
//	vflags |= X509_V_FLAG_ALLOW_PROXY_CERTS;
//	vflags &= !X509_V_FLAG_X509_STRICT;
	X509_STORE_set_flags(store, vflags);

	if (!(verify_ctx = X509_STORE_CTX_new())) {
		LOG(L_ERROR, "failed to create store ctx");
		return 6;
	}

	X509_NAME_oneline(X509_get_subject_name(cert), cert_dn, sizeof(cert_dn));
	X509_NAME_oneline(X509_get_issuer_name(cert), issuer_dn, sizeof(issuer_dn));

	LOG(L_DEBUG, "validating DN(%s), ISSUER(%s)", cert_dn, issuer_dn);

	if (X509_STORE_CTX_init(verify_ctx, store, cert, cert_chain) != 1) {
		LOG(L_ERROR, "failed to init store ctx");
		return 7;
	}

	X509_STORE_CTX_set_purpose(verify_ctx, X509_PURPOSE_SSL_CLIENT);

	if (X509_verify_cert(verify_ctx) != 1) {
		LOG(L_ERROR, "chain not valid (%d:%s)", verify_ctx->error,
				X509_verify_cert_error_string(verify_ctx->error));
		return 8;
	} else {
		LOG(L_DEBUG, "chain is valid");
	}

	return 0;
}


static unsigned long verify_proxy(STACK_OF(X509) *chain, int start_pos,
		int end_pos) {
   int      i = 0;
    X509    *cert = NULL, *issuer = NULL;
    time_t   now = time((time_t *)NULL);
    size_t   len = 0;             /* Lengths of issuer and cert DN */
    size_t   len2 = 0;            /* Lengths of issuer and cert DN */
    int      prev_is_limited = 0;   /* previous cert was proxy and limited */
    char    *proxy_part_dn = NULL;
    int      depth;
    int      is_old_style_proxy = 0;
    int      is_limited_proxy = 0;
    ASN1_INTEGER   *cert_serial = NULL;
    ASN1_INTEGER   *issuer_serial = NULL;
	char subject_dn[256], issuer_dn[256], issuer_subject_dn[256];

    /* And there was (current) time... */
    time(&now);

	issuer = NULL;

    depth = sk_X509_num(chain);
	
    /* Changed this value to start checking the proxy and such and
       to skip the CA and the user_cert
    */
	for (i = start_pos; i < end_pos && i < depth - 1; ++i) {
        /* Check for X509 certificate and point to it with 'cert' */
		cert = sk_X509_value(chain, i);

		if (i + 1 < depth) {
			issuer = sk_X509_value(chain, i + 1);
		} else {
			LOG(L_ERROR, "proxy certificate doesn't contain issuer (%d vs %d)",
					i + 1, depth);
			return X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY;
		}


        X509_NAME_oneline(X509_get_subject_name(cert), subject_dn,
				sizeof(subject_dn));
        X509_NAME_oneline(X509_get_issuer_name(cert), issuer_dn,
				sizeof(issuer_dn));
        X509_NAME_oneline(X509_get_subject_name(issuer), issuer_subject_dn,
				sizeof(issuer_subject_dn));

        len  = strlen(subject_dn);
        len2 = strlen(issuer_dn);

        LOG(L_DEBUG, "verifying proxy [%s] with issuer [%s]", subject_dn,
				issuer_dn);
   
		if (strcmp(issuer_subject_dn, issuer_dn)) {
			LOG(L_ERROR, "failed to find issuer");
			return X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY;
		}

        if (now < grid_asn1TimeToTimeT((char*)ASN1_STRING_data(
						X509_get_notBefore(cert)),0)) {
                LOG(L_ERROR, "proxy certificate is not yet valid" );
                return X509_V_ERR_CERT_NOT_YET_VALID;
        }

        if (now > grid_asn1TimeToTimeT((char*)ASN1_STRING_data(
						X509_get_notAfter(cert)), 0)) {
              LOG(L_ERROR, "proxy certificate expired");
                return X509_V_ERR_CERT_HAS_EXPIRED;
		}

		/* User not allowed to sign shortened DN */
		if (len2 > len) {
			LOG(L_ERROR, "it is not allowed to sign a shorthened DN");
			return X509_V_ERR_APPLICATION_VERIFICATION;
		}

		/* Proxy subject must begin with issuer. */
		if (strncmp(subject_dn, issuer_dn, len2) != 0) {
			LOG(L_ERROR, "proxy subject must begin with the issuer");
			return X509_V_ERR_APPLICATION_VERIFICATION;
		}

		/* Set pointer to end of base DN in cert_DN */
		proxy_part_dn = &subject_dn[len2];

		/* First attempt at support for Old and New style GSI
		   proxies: /CN=anything is ok for now */
		if (strncmp(proxy_part_dn, "/CN=", 4) != 0) {
			LOG(L_ERROR, "could not find a /CN= structure in the DN, "
					"thus it is not a proxy");
			return X509_V_ERR_APPLICATION_VERIFICATION;
		}

		if (strncmp(proxy_part_dn, "/CN=proxy", 9) == 0) {
			LOG(L_DEBUG, "current certificate is an old style proxy");
			is_old_style_proxy = 1;
			is_limited_proxy = 0;
		} else if (strncmp(proxy_part_dn, "/CN=limited proxy", 17) == 0) {
			LOG(L_DEBUG, "current certificate is an old limited style proxy");
			is_old_style_proxy = 1;
			is_limited_proxy = 1;
		} else {
			LOG(L_DEBUG, "current certificate is a GSI/RFC3820 proxy");
		}
  
		if (is_old_style_proxy) {
			cert_serial = X509_get_serialNumber(cert);
			issuer_serial = X509_get_serialNumber(issuer);

			if (cert_serial && issuer_serial) {
				if (ASN1_INTEGER_cmp(cert_serial, issuer_serial)) {
					LOG(L_DEBUG, "serial numbers do not match");
					//return -2;
				}
			}
		}

		if (is_limited_proxy) {
			prev_is_limited = 1;
			if (i > 0)
				LOG(L_DEBUG, "found limited proxy");
		} else {
			if (prev_is_limited) {
				LOG(L_ERROR, "proxy chain integrity error. Previous proxy "
						"in chain was limited, but this one is a regular proxy");
				return X509_V_ERR_APPLICATION_VERIFICATION;
			}
		}
	}

    return X509_V_OK;
}


static time_t grid_asn1TimeToTimeT(char *asn1time, size_t len) {
   char   zone;
   struct tm time_tm;
   
   if (len == 0) len = strlen(asn1time);
                                                                                
   if ((len != 13) && (len != 15)) return 0; /* dont understand */
                                                                                
   if ((len == 13) &&
       ((sscanf(asn1time, "%02d%02d%02d%02d%02d%02d%c",
         &(time_tm.tm_year),
         &(time_tm.tm_mon),
         &(time_tm.tm_mday),
         &(time_tm.tm_hour),
         &(time_tm.tm_min),
         &(time_tm.tm_sec),
         &zone) != 7) || (zone != 'Z'))) return 0; /* dont understand */
                                                                                
   if ((len == 15) &&
       ((sscanf(asn1time, "20%02d%02d%02d%02d%02d%02d%c",
         &(time_tm.tm_year),
         &(time_tm.tm_mon),
         &(time_tm.tm_mday),
         &(time_tm.tm_hour),
         &(time_tm.tm_min),
         &(time_tm.tm_sec),
         &zone) != 7) || (zone != 'Z'))) return 0; /* dont understand */
                                                                                
   /* time format fixups */
                                                                                
   if (time_tm.tm_year < 90) time_tm.tm_year += 100;
   --(time_tm.tm_mon);
                                                                                
   return timegm(&time_tm);
}


/*****************************************************************************/

